# -*- coding: utf-8 -*-

"""
Copyright 2016 William La Cava

This file is part of the FEW library.

License: GPLv3
"""

from ._version import __version__
from .few import FEW, main
# from .few import evaluation
# from .few import selection
# from .few import variation
